"""
    Logic to provide consistency across exchanges
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import time

import pandas as pd

import blankly.utils.exceptions as exceptions
import blankly.utils.utils
import blankly.utils.utils as utils
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.exchanges.orders.stop_loss import StopLossOrder
from blankly.exchanges.orders.take_profit import TakeProfitOrder


class BinanceInterface(ExchangeInterface):
    _asset_precision: dict

    def __init__(self, exchange_name, authenticated_api):
        self._asset_precision = {}
        # Initialize this as None so that it can be filled & cached when needed
        self.__available_currencies = None
        super().__init__(exchange_name, authenticated_api, valid_resolutions=[60, 180, 300, 900, 1800, 3600, 7200,
                                                                              14400, 21600, 28800, 43200, 86400, 259200,
                                                                              604800, 2592000])

    def init_exchange(self):
        import binance.exceptions
        try:
            self.calls.get_account()
        except binance.exceptions.BinanceAPIException as e:
            raise exceptions.APIException(f"Debugging info: {e} - {e.response} - {e.message}\n"
                                          "Invalid API Key, IP, or permissions for action - are you trying "
                                          "to use your normal exchange keys while in sandbox mode? "
                                          "\nTry toggling the \'sandbox\' setting in your keys.json, check "
                                          "if the keys were input correctly into your keys.json or ensure you have set "
                                          "the correct binance_tld in settings.json.\n")

        symbols = self.calls.get_exchange_info()["symbols"]
        assets = []
        for i in symbols:
            assets.append(i["baseAsset"])
            assets.append(i["quoteAsset"])

        # Because these come down as trading pairs we have to filter for duplicates
        filtered_base_assets = []
        for i in assets:
            # TODO replace this with a lambda function using the filter method
            if i not in filtered_base_assets:
                filtered_base_assets.append(i)
        self.__available_currencies = filtered_base_assets

    def get_products(self):
        needed = self.needed['get_products']
        """
        This is a section of the symbols array
        [
            {
                "symbol": "BTCUSD",
                "status": "TRADING",
                "baseAsset": "BTC",
                "baseAssetPrecision": 8,
                "quoteAsset": "USD",
                "quotePrecision": 4,
                "quoteAssetPrecision": 4,
                "baseCommissionPrecision": 8,
                "quoteCommissionPrecision": 2,
                "orderTypes": [
                    "LIMIT",
                    "LIMIT_MAKER",
                    "MARKET",
                    "STOP_LOSS_LIMIT",
                    "TAKE_PROFIT_LIMIT",
                ],
                "icebergAllowed": True,
                "ocoAllowed": True,
                "quoteOrderQtyMarketAllowed": True,
                "isSpotTradingAllowed": True,
                "isMarginTradingAllowed": False,
                "filters": [
                    {
                        "filterType": "PRICE_FILTER",
                        "minPrice": "0.0100",
                        "maxPrice": "100000.0000",
                        "tickSize": "0.0100",
                    },
                    {
                        "filterType": "PERCENT_PRICE",
                        "multiplierUp": "5",
                        "multiplierDown": "0.2",
                        "avgPriceMins": 5,
                    },
                    {
                        "filterType": "LOT_SIZE",
                        "minQty": "0.00000100",
                        "maxQty": "9000.00000000",
                        "stepSize": "0.00000100",
                    },
                    {
                        "filterType": "MIN_NOTIONAL",
                        "minNotional": "10.0000",
                        "applyToMarket": True,
                        "avgPriceMins": 5,
                    },
                    {"filterType": "ICEBERG_PARTS", "limit": 10},
                    {
                        "filterType": "MARKET_LOT_SIZE",
                        "minQty": "0.00000000",
                        "maxQty": "3200.00000000",
                        "stepSize": "0.00000000",
                    },
                    {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": 200},
                    {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": 5},
                ],
                "permissions": ["SPOT"],
            },
        ]
        """
        products = self.calls.get_exchange_info()["symbols"]
        for i in range(len(products)):
            # Rename needed
            products[i]["symbol"] = products[i]["baseAsset"] + "-" + products[i]["quoteAsset"]
            products[i]["base_asset"] = products[i].pop("baseAsset")
            products[i]["quote_asset"] = products[i].pop("quoteAsset")
            filters = products[i]["filters"]
            # Iterate to find the next few
            min_qty = None
            max_qty = None
            base_increment = None
            for filters_array in filters:
                if filters_array["filterType"] == "LOT_SIZE":
                    min_qty = filters_array["minQty"]
                    max_qty = filters_array["maxQty"]
                    base_increment = filters_array["stepSize"]
                    break
            products[i]["base_min_size"] = min_qty
            products[i]["base_max_size"] = max_qty
            products[i]["base_increment"] = base_increment
            # Isolate keys unimportant for the interface's functionality
            products[i] = utils.isolate_specific(needed, products[i])
        return products

    @utils.enforce_base_asset
    def get_account(self, symbol=None) -> utils.AttributeDict:
        """
        Get all currencies in an account, or sort by asset/account_id
        Args:
            symbol (Optional): Filter by particular asset

            These arguments are mutually exclusive
        Coinbase Pro: get_account
        binance: get_account["balances"]
        """
        symbol = super().get_account(symbol=symbol)

        # TODO this should really use the get_asset_balance() function from binance.
        """
        {
            "makerCommission": 15,
            "takerCommission": 15,
            "buyerCommission": 0,
            "sellerCommission": 0,
            "canTrade": true,
            "canWithdraw": true,
            "canDeposit": true,
            "balances": [
                {
                    "asset": "BTC",
                    "free": "4723846.89208129",
                    "locked": "0.00000000"
                },
                {
                    "asset": "LTC",
                    "free": "4763368.68006011",
                    "locked": "0.00000000"
                }
            ]
        }
        """
        renames = [
            ["asset", "symbol"],
            ["free", "available"],
            ["locked", "hold"],
        ]

        # Note that it's now a possibility that the available currencies were never downloaded, so we can do that now
        if self.__available_currencies is None:
            self.init_exchange()

        accounts = self.calls.get_account()["balances"]

        # Isolate for symbol, warn if not found or default to just returning a parsed version
        if symbol is not None:
            if symbol in self.__available_currencies:
                for i in range(len(accounts)):
                    # If it was in the accounts return we can just isolate & parse
                    if accounts[i]["asset"] == symbol:
                        accounts = utils.rename_to(renames, accounts[i])
                        return utils.AttributeDict({
                            'available': float(accounts['available']),
                            'hold': float(accounts['hold'])
                        })
                # If not just return a default 0 value. This is safe because we already checked if the symbol
                #  was valid
                return utils.AttributeDict({
                    "available": 0.0,
                    "hold": 0.0
                })
            else:
                raise LookupError("Symbol was not found")

        # binance only returns things you own, so extract those - scan and add that to the array.
        # We can fill it to a balance of zero later
        owned_assets = []
        for i in accounts:
            owned_assets.append(i['asset'])
        # Create an empty list for return
        parsed_dictionary = {}
        # Iterate through
        for i in self.__available_currencies:
            # Iterate through everything binance returned
            found = False
            for val in accounts:
                # If the current available asset matches one from binance
                if val['asset'] == i:
                    # Do the normal thing above and append
                    mutated = utils.rename_to(renames, val)
                    parsed_dictionary[mutated['symbol']] = utils.AttributeDict({
                        'available': float(mutated['available']),
                        'hold': float(mutated['hold'])
                    })
                    found = True
            # If it wasn't found just default here
            if not found:
                parsed_dictionary[i] = utils.AttributeDict({
                    'available': 0.0,
                    'hold': 0.0
                })
        return utils.AttributeDict(parsed_dictionary)

    def _fix_response(self, needed, response):
        response['side'] = response['side'].lower()
        response['type'] = response['type'].lower()
        response['status'] = super().homogenize_order_status('binance', response['status'].lower())
        response['symbol'] = utils.to_blankly_symbol(response['symbol'], 'binance')
        if 'transactTime' in response:
            response["transactTime"] /= 1000
        response = utils.rename_to([
            ["orderId", "id"],
            ["transactTime", "created_at"],
            ["origQty", "size"],
            ["timeInForce", "time_in_force"],
            ["cummulativeQuoteQty", "funds"]
        ], response)
        response = utils.isolate_specific(needed, response)
        return response

    @utils.order_protection
    def market_order(self, symbol, side, size) -> MarketOrder:
        """
        Used for buying or selling market orders
        Args:
            symbol: asset to buy
            side: buy/sell
            size: desired amount of base asset to use
        """
        needed = self.needed['market_order']
        """
        Response RESULT:

        .. code-block:: python

        {
            "symbol": "BTCUSDT",        <-- Similar
            "orderId": 28,      <-- Similar
            "clientOrderId": "6gCrw2kRUAF9CvJDGP16IP",
            "transactTime": 1507725176595,      <-- Similar
            "price": "0.00000000",      <-- Similar
            "origQty": "10.00000000",       <-- Similar
            "executedQty": "10.00000000",
            "status": "FILLED",     <-- Similar
            "timeInForce": "GTC",       <-- Similar
            "type": "MARKET",       <-- Similar
            "side": "SELL"      <-- Similar
        }
        
        Actual response buying 20 dollars worth
        {
            'symbol': 'BTCUSDT', 
            'orderId': 3038554, 
            'orderListId': -1, 
            'clientOrderId': 'xJqha6DQgdYHNdxuVqV2BZ', 
            'transactTime': 1623800280187, 
            'price': '0.00000000', 
            'origQty': '0.00049600', 
            'executedQty': '0.00049600', 
            'cummulativeQuoteQty': '19.99128000', 
            'status': 'FILLED', 
            'timeInForce': 'GTC', 
            'type': 'MARKET', 
            'side': 'BUY', 
            'fills': [
                {
                    'price': '40305.00000000', 
                    'qty': '0.00049600', 
                    'commission': '0.00000000', 
                    'commissionAsset': 'BTC', 
                    'tradeId': 564349
                }
            ]
        }
        """
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = {
            'size': size,
            'side': side,
            'symbol': symbol,
            'type': 'market'
        }
        modified_symbol = utils.to_exchange_symbol(symbol, "binance")
        # The interface here will be the query of order status from this object, because orders are dynamic
        # creatures
        response = self.calls.order_market(symbol=modified_symbol, side=side, quantity=size)
        response = self._fix_response(needed, response)
        return MarketOrder(order, response, self)

    @utils.order_protection
    def limit_order(self, symbol, side, price, size) -> LimitOrder:
        """
        Used for buying or selling limit orders
        Args:
            symbol: currency to buy
            side: buy/sell
            price: price to set limit order
            size: amount of currency (like BTC) for the limit to be valued
        """
        needed = self.needed['limit_order']
        """Send in a new limit order

        Any order with an icebergQty MUST have timeInForce set to GTC.

        :param symbol: required
        :type symbol: str
        :param side: required
        :type side: str
        :param quantity: required
        :type quantity: decimal
        :param price: required
        :type price: str
        :param timeInForce: default Good till cancelled
        :type timeInForce: str
        :param newClientOrderId: A unique id for the order. Automatically generated if not sent.
        :type newClientOrderId: str
        :param icebergQty: Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        :type icebergQty: decimal
        :param newOrderRespType: Set the response JSON. ACK, RESULT, or FULL; default: RESULT.
        :type newOrderRespType: str
        :param recvWindow: the number of milliseconds the request is valid for
        :type recvWindow: int

        :returns: API response

        See order endpoint for full response options

        :raises: BinanceRequestException, BinanceAPIException, BinanceOrderException, BinanceOrderMinAmountException,
        BinanceOrderMinPriceException, BinanceOrderMinTotalException,
        BinanceOrderUnknownSymbolException, BinanceOrderInactiveSymbolException

        """
        if self.should_auto_trunc:
            size = utils.trunc(size, self.get_asset_precision(symbol))
        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'limit'
        }
        modified_symbol = utils.to_exchange_symbol(symbol, 'binance')
        response = self.calls.order_limit(symbol=modified_symbol, side=side, price=price, quantity=size)
        response = self._fix_response(needed, response)
        return LimitOrder(order, response, self)

    @utils.order_protection
    def take_profit_order(self, symbol, price, size) -> TakeProfitOrder:
        """
        Used for sending take profit orders
        Args:
            symbol: currency to sell
            price: price to sell at
            size: amount of currency (like BTC)
        """
        needed = self.needed['take_profit']
        """Send in a new take-profit order

        Any order with an icebergQty MUST have timeInForce set to GTC.

        :param symbol: required
        :type symbol: str
        :param side: required
        :type side: str
        :param quantity: required
        :type quantity: decimal
        :param price: required
        :type price: str
        :param timeInForce: default Good till cancelled
        :type timeInForce: str
        :param newClientOrderId: A unique id for the order. Automatically generated if not sent.
        :type newClientOrderId: str
        :param icebergQty: Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        :type icebergQty: decimal
        :param newOrderRespType: Set the response JSON. ACK, RESULT, or FULL; default: RESULT.
        :type newOrderRespType: str
        :param recvWindow: the number of milliseconds the request is valid for
        :type recvWindow: int

        :returns: API response

        See order endpoint for full response options

        :raises: BinanceRequestException, BinanceAPIException, BinanceOrderException, BinanceOrderMinAmountException,
        BinanceOrderMinPriceException, BinanceOrderMinTotalException,
        BinanceOrderUnknownSymbolException, BinanceOrderInactiveSymbolException

        """
        side = 'sell'
        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'take_profit'
        }
        modified_symbol = utils.to_exchange_symbol(symbol, 'binance')
        response = self.calls.create_order(symbol=modified_symbol, side=side, stopPrice=price, quantity=size,
                                           type='TAKE_PROFIT')
        response = self._fix_response(needed, response)
        return TakeProfitOrder(order, response, self)

    @utils.order_protection
    def stop_loss_order(self, symbol, price, size) -> StopLossOrder:
        """
        Used for sending stop loss orders
        Args:
            symbol: currency to sell
            price: price to sell at
            size: amount of currency (like BTC)
        """
        needed = self.needed['stop_loss']
        """Send in a new stop-loss order

        Any order with an icebergQty MUST have timeInForce set to GTC.

        :param symbol: required
        :type symbol: str
        :param side: required
        :type side: str
        :param quantity: required
        :type quantity: decimal
        :param price: required
        :type price: str
        :param timeInForce: default Good till cancelled
        :type timeInForce: str
        :param newClientOrderId: A unique id for the order. Automatically generated if not sent.
        :type newClientOrderId: str
        :param icebergQty: Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        :type icebergQty: decimal
        :param newOrderRespType: Set the response JSON. ACK, RESULT, or FULL; default: RESULT.
        :type newOrderRespType: str
        :param recvWindow: the number of milliseconds the request is valid for
        :type recvWindow: int

        :returns: API response

        See order endpoint for full response options

        :raises: BinanceRequestException, BinanceAPIException, BinanceOrderException, BinanceOrderMinAmountException,
        BinanceOrderMinPriceException, BinanceOrderMinTotalException,
        BinanceOrderUnknownSymbolException, BinanceOrderInactiveSymbolException

        """
        side = 'sell'
        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'stop_loss'
        }
        modified_symbol = utils.to_exchange_symbol(symbol, 'binance')
        response = self.calls.create_order(symbol=modified_symbol, side=side, stopPrice=price, quantity=size,
                                           type='STOP_LOSS')
        response = self._fix_response(needed, response)
        return StopLossOrder(order, response, self)

    def cancel_order(self, symbol, order_id) -> dict:
        """Cancel an active order. Either orderId or origClientOrderId must be sent.

        https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#cancel-order-trade

        :returns: API response

        .. code-block:: python

        {
          "symbol": "LTCBTC",
          "origClientOrderId": "myOrder1",
          "orderId": 4,
          "orderListId": -1, //Unless part of an OCO, the value will always be -1.
          "clientOrderId": "cancelMyOrder1",
          "price": "2.00000000",
          "origQty": "1.00000000",
          "executedQty": "0.00000000",
          "cummulativeQuoteQty": "0.00000000",
          "status": "CANCELED",
          "timeInForce": "GTC",
          "type": "LIMIT",
          "side": "BUY"
        }

        :raises: BinanceRequestException, BinanceAPIException

        """
        needed = self.needed['cancel_order']
        renames = [
            ["orderId", "order_id"]
        ]
        modified_symbol = utils.to_exchange_symbol(symbol, 'binance')
        response = self.calls.cancel_order(symbol=modified_symbol, orderId=order_id)
        response = utils.rename_to(renames, response)
        return utils.isolate_specific(needed, response)

    def get_open_orders(self, symbol=None):
        """
        List open orders.
        """
        """
        [
            {
                "symbol": "LTCBTC",
                "orderId": 1,
                "orderListId": -1, //Unless OCO, the value will always be -1
                "clientOrderId": "myOrder1",
                "price": "0.1",
                "origQty": "1.0",    <- what are these??
                "executedQty": "0.0",    <- what are these??
                "cummulativeQuoteQty": "0.0",    <- what are these??
                "status": "NEW",
                "timeInForce": "GTC",
                "type": "LIMIT",
                "side": "BUY",
                "stopPrice": "0.0",
                "icebergQty": "0.0",
                "time": 1499827319559,
                "updateTime": 1499827319559,
                "isWorking": true,
                "origQuoteOrderQty": "0.000000"    <- what are these??
            }
        ]
        """
        renames = [
            ["orderId", "id"],
            ["origQty", "size"],
            ["cummulativeQuoteQty", "funds"],
            ["time", "created_at"],
            ["timeInForce", "time_in_force"]
        ]

        if symbol is not None:
            symbol = utils.to_exchange_symbol(symbol, "binance")
            orders = self.calls.get_open_orders(symbol=symbol)
        else:
            orders = self.calls.get_open_orders()

        for i in range(len(orders)):
            orders[i] = utils.rename_to(renames, orders[i])
            orders[i]['type'] = orders[i]['type'].lower()
            orders[i]['side'] = orders[i]['side'].lower()
            orders[i]['status'] = super().homogenize_order_status('binance', orders[i]['status'].lower())
            orders[i]['created_at'] = orders[i]['created_at'] / 1000
            needed = self.choose_order_specificity(orders[i]['type'])
            orders[i] = utils.isolate_specific(needed, orders[i])
            orders[i]['symbol'] = utils.to_blankly_symbol(orders[i]['symbol'], 'binance', quote_guess=None)

        return orders

    def get_order(self, symbol, order_id) -> dict:
        """
        {
            "symbol": "LTCBTC",
            "orderId": 1,
            "clientOrderId": "myOrder1",
            "price": "0.1",
            "origQty": "1.0",
            "executedQty": "0.0",
            "status": "NEW",
            "timeInForce": "GTC",
            "type": "LIMIT",
            "side": "BUY",
            "stopPrice": "0.0",
            "icebergQty": "0.0",
            "time": 1499827319559
        }
        """
        renames = [
            ["orderId", "id"],
            ["origQty", "size"],
            ["cummulativeQuoteQty", "funds"],
            ["time", "created_at"],
            ["timeInForce", "time_in_force"]
        ]

        symbol = utils.to_exchange_symbol(symbol, 'binance')
        response = self.calls.get_order(symbol=symbol, orderId=order_id)

        response = utils.rename_to(renames, response)
        response['type'] = response['type'].lower()
        response['side'] = response['side'].lower()
        response['status'] = super().homogenize_order_status('binance', response['status'].lower())
        response['created_at'] = response['created_at'] / 1000
        response['symbol'] = utils.to_blankly_symbol(response['symbol'], 'binance')
        needed = self.choose_order_specificity(response['type'])
        response = utils.isolate_specific(needed, response)

        return response

    """
    Coinbase Pro: get_fees
    binance: get_trade_fee
    """

    def get_fees(self, symbol) -> dict:
        needed = self.needed['get_fees']
        """
        {
            "makerCommission": 15,
            "takerCommission": 15,
            "buyerCommission": 0,
            "sellerCommission": 0,
            "canTrade": true,
            "canWithdraw": true,
            "canDeposit": true,
            "balances": [
                {
                    "asset": "BTC",
                    "free": "4723846.89208129",
                    "locked": "0.00000000"
                },
                {
                    "asset": "LTC",
                    "free": "4763368.68006011",
                    "locked": "0.00000000"
                }
            ]
        }
        """
        # TODO: make sure this supports with binance Coin (BNB) discount
        account = self.calls.get_account()
        # Get rid of the stuff we really don't need this time
        account.pop('balances')
        # Rename makers and takers
        account['maker_fee_rate'] = account.pop('makerCommission') / 10000
        account['taker_fee_rate'] = account.pop('takerCommission') / 10000
        # Isolate
        return utils.isolate_specific(needed, account)

    """
    Coinbase Pro: get_account_history
    binance:
        get_deposit_history
        get_withdraw_history

    """

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        """
        Returns the product history from an exchange
        Args:
            symbol: blankly product ID format (BTC-USD)
            epoch_start: Time to begin download
            epoch_stop: Time to stop download
            resolution: Resolution in seconds between tick (ex: 60 = 1 per minute)
        Returns:
            Dataframe with *at least* 'time (epoch)', 'low', 'high', 'open', 'close', 'volume' as columns.
        """
        return self._binance_get_product_history(self.calls, symbol, epoch_start, epoch_stop, resolution)

    @staticmethod
    def _binance_get_product_history(calls, symbol, epoch_start, epoch_stop, resolution):
        resolution = blankly.time_builder.time_interval_to_seconds(resolution)

        # epoch_start, epoch_stop = super().get_product_history(symbol, epoch_start, epoch_stop, resolution)
        epoch_start = utils.convert_epochs(epoch_start)
        epoch_stop = utils.convert_epochs(epoch_stop)

        epoch_start = int(epoch_start)
        epoch_stop = int(epoch_stop)

        accepted_grans = [60, 180, 300, 900, 1800, 3600, 7200, 14400,
                          21600, 28800, 43200, 86400, 259200, 604800, 2592000]
        if resolution not in accepted_grans:
            utils.info_print("Granularity is not an accepted granularity...rounding to nearest valid value.")
            resolution = accepted_grans[min(range(len(accepted_grans)),
                                            key=lambda i: abs(accepted_grans[i] - resolution))]
        lookup_dict = {
            60: "1m",
            180: "3m",
            300: "5m",
            900: "15m",
            1800: "30m",
            3600: "1h",
            7200: "2h",
            14400: "4h",
            21600: "6h",
            28800: "8h",
            43200: "12h",
            86400: "1d",
            259200: "3d",
            604800: "1w",
            2592000: "1M"
        }
        gran_string = lookup_dict[resolution]

        # Figure out how many points are needed
        need = int((epoch_stop - epoch_start) / resolution)
        initial_need = need
        window_open = epoch_start
        history = []

        # Convert coin id to binance coin
        symbol = utils.to_exchange_symbol(symbol, 'binance')
        while need > 1000:
            # Close is always 300 points ahead
            window_close = int(window_open + 1000 * resolution)
            history = history + calls.get_klines(symbol=symbol, startTime=window_open * 1000,
                                                 endTime=window_close * 1000, interval=gran_string,
                                                 limit=1000)

            window_open = window_close
            need -= 1000
            time.sleep(.2)
            utils.update_progress((initial_need - need) / initial_need)

        # Fill the remainder
        history_block = history + calls.get_klines(symbol=symbol, startTime=window_open * 1000,
                                                   endTime=epoch_stop * 1000, interval=gran_string,
                                                   limit=1000)

        data_frame = pd.DataFrame(history_block, columns=['time', 'open', 'high', 'low', 'close', 'volume',
                                                          'close time', 'quote asset volume', 'number of trades',
                                                          'taker buy base asset volume',
                                                          'taker buy quote asset volume', 'ignore'], dtype=None)
        # Clear the ignore column, why is that there binance?
        del data_frame['ignore']

        # Want them in this order: ['time (epoch)', 'low', 'high', 'open', 'close', 'volume']

        # Time is so big it has to be cast separately for windows
        data_frame['time'] = data_frame['time'].div(1000).astype(int)

        # Cast dataframe
        data_frame = data_frame.astype({
            'open': float,
            'high': float,
            'low': float,
            'close': float,
            'volume': float,
            'close time': int,
            'quote asset volume': float,
            'number of trades': int,
            'taker buy base asset volume': float,
            'taker buy quote asset volume': float
        })

        # Convert time to seconds
        return data_frame.reindex(columns=['time', 'low', 'high', 'open', 'close', 'volume'])

    """
    Coinbase Pro: Get Currencies
    binance: get_products
    """

    def get_order_filter(self, symbol):
        """
        Optimally we'll just remove the filter section and make the returns accurate


        FOR THE US EXCHANGE:
        {
            "symbol": "BTCUSD",
            "status": "TRADING",
            "baseAsset": "BTC",
            "baseAssetPrecision": 8,
            "quoteAsset": "USD",
            "quotePrecision": 4,
            "quoteAssetPrecision": 4,
            "baseCommissionPrecision": 8,
            "quoteCommissionPrecision": 2,
            "orderTypes": [
                "LIMIT",
                "LIMIT_MAKER",
                "MARKET",
                "STOP_LOSS_LIMIT",
                "TAKE_PROFIT_LIMIT",
            ],
            "icebergAllowed": True,
            "ocoAllowed": True,
            "quoteOrderQtyMarketAllowed": True,
            "isSpotTradingAllowed": True,
            "isMarginTradingAllowed": False,
            "filters": [
                {
                    "filterType": "PRICE_FILTER",
                    "minPrice": "0.0100",
                    "maxPrice": "100000.0000",
                    "tickSize": "0.0100",
                },
                {
                    "filterType": "PERCENT_PRICE",
                    "multiplierUp": "5",
                    "multiplierDown": "0.2",
                    "avgPriceMins": 5,
                },
                {
                    "filterType": "LOT_SIZE",
                    "minQty": "0.00000100",
                    "maxQty": "9000.00000000",
                    "stepSize": "0.00000100",
                },
                {
                    "filterType": "MIN_NOTIONAL",
                    "minNotional": "10.0000",
                    "applyToMarket": True,
                    "avgPriceMins": 5,
                },
                {"filterType": "ICEBERG_PARTS", "limit": 10},
                {
                    "filterType": "MARKET_LOT_SIZE",
                    "minQty": "0.00000000",
                    "maxQty": "3200.00000000",
                    "stepSize": "0.00000000",
                },
                {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": 200},
                {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": 5},
            ],
            "permissions": ["SPOT"],
        }


        FOR THE INTERNATIONAL EXCHANGE:
        {
            "symbol": "ETHUSDT",
            "status": "TRADING",
            "baseAsset": "ETH",
            "baseAssetPrecision": 8,
            "quoteAsset": "USDT",
            "quotePrecision": 8,
            "quoteAssetPrecision": 8,
            "baseCommissionPrecision": 8,
            "quoteCommissionPrecision": 8,
            "orderTypes":
            [
                "LIMIT",
                "LIMIT_MAKER",
                "MARKET",
                "STOP_LOSS_LIMIT",
                "TAKE_PROFIT_LIMIT"
            ],
            "icebergAllowed": true,
            "ocoAllowed": true,
            "quoteOrderQtyMarketAllowed": true,
            "allowTrailingStop": true,
            "cancelReplaceAllowed": true,
            "isSpotTradingAllowed": true,
            "isMarginTradingAllowed": true,
            "filters":
            [
                {
                    "filterType": "PRICE_FILTER",
                    "minPrice": "0.01000000",
                    "maxPrice": "1000000.00000000",
                    "tickSize": "0.01000000"
                },
                {
                    "filterType": "LOT_SIZE",
                    "minQty": "0.00010000",
                    "maxQty": "9000.00000000",
                    "stepSize": "0.00010000"
                },
                {
                    "filterType": "MIN_NOTIONAL",
                    "minNotional": "10.00000000",
                    "applyToMarket": true,
                    "avgPriceMins": 5
                },
                {
                    "filterType": "ICEBERG_PARTS",
                    "limit": 10
                },
                {
                    "filterType": "MARKET_LOT_SIZE",
                    "minQty": "0.00000000",
                    "maxQty": "4382.40354798",
                    "stepSize": "0.00000000"
                },
                {
                    "filterType": "TRAILING_DELTA",
                    "minTrailingAboveDelta": 10,
                    "maxTrailingAboveDelta": 2000,
                    "minTrailingBelowDelta": 10,
                    "maxTrailingBelowDelta": 2000
                },
                {
                    "filterType": "PERCENT_PRICE_BY_SIDE",
                    "bidMultiplierUp": "5",
                    "bidMultiplierDown": "0.2",
                    "askMultiplierUp": "5",
                    "askMultiplierDown": "0.2",
                    "avgPriceMins": 5
                },
                {
                    "filterType": "MAX_NUM_ORDERS",
                    "maxNumOrders": 200
                },
                {
                    "filterType": "MAX_NUM_ALGO_ORDERS",
                    "maxNumAlgoOrders": 5
                }
            ],
            "permissions":
            [
                "SPOT",
                "MARGIN",
                "TRD_GRP_004",
                "TRD_GRP_005"
            ],
            "defaultSelfTradePreventionMode": "NONE",
            "allowedSelfTradePreventionModes":
            [
                "NONE"
            ]
        }
        """
        us_filter_mapping = {
            'PRICE_FILTER': 0,
            'PERCENT_PRICE': 1,
            'LOT_SIZE': 2,
            'MIN_NOTIONAL': 3,
            'ICEBERG_PARTS': 4,
            'MARKET_LOT_SIZE': 5,
            'TRAILING_DELTA': 6,
            'MAX_NUM_ORDERS': 7,
            'MAX_NUM_ALGO_ORDERS': 8
        }

        international_filter_mapping = {
            'PRICE_FILTER': 0,
            'LOT_SIZE': 1,
            'MIN_NOTIONAL': 2,
            'ICEBERG_PARTS': 3,
            'MARKET_LOT_SIZE': 4,
            'TRAILING_DELTA': 5,
            'PERCENT_PRICE_BY_SIDE': 6,
            'MAX_NUM_ORDERS': 7,
            'MAX_NUM_ALGO_ORDERS': 8
        }

        def is_international_exchange(asset_dict: dict) -> bool:
            """
            Given the dictionary for the asset, tell if it is an international exchange
            """
            filters_ = asset_dict['filters']

            return filters_[1]['filterType'] == 'LOT_SIZE'

        converted_symbol = utils.to_exchange_symbol(symbol, 'binance')
        current_price = None
        symbol_data = self.calls.get_exchange_info()["symbols"]
        for i in symbol_data:
            if i["symbol"] == converted_symbol:
                symbol_data = i
                current_price = float(self.calls.get_avg_price(symbol=converted_symbol)['price'])
                break


        if current_price is None:
            raise LookupError("Specified market not found")

        using_international_exchange = is_international_exchange(symbol_data)
        if using_international_exchange:
            filter_mapping = international_filter_mapping
        else:
            filter_mapping = us_filter_mapping

        filters = symbol_data["filters"]
        hard_min_price = float(filters[filter_mapping['PRICE_FILTER']]["minPrice"])
        hard_max_price = float(filters[filter_mapping['PRICE_FILTER']]["maxPrice"])
        quote_increment = float(filters[filter_mapping['PRICE_FILTER']]["tickSize"])

        # This is the only one where the keys are different
        if using_international_exchange:
            # TODO technically this multiplier is different for buy and sell sides, this should be reflected in the
            #  backtesting engine
            multiplier_up = float(filters[filter_mapping['PERCENT_PRICE_BY_SIDE']]["bidMultiplierUp"])
            multiplier_down = float(filters[filter_mapping['PERCENT_PRICE_BY_SIDE']]["bidMultiplierDown"])
        else:
            multiplier_up = float(filters[filter_mapping['PERCENT_PRICE']]["multiplierUp"])
            multiplier_down = float(filters[filter_mapping['PERCENT_PRICE']]["multiplierDown"])

        percent_min_price = multiplier_down * current_price
        percent_max_price = multiplier_up * current_price

        min_quantity = float(filters[filter_mapping['LOT_SIZE']]["minQty"])
        max_quantity = float(filters[filter_mapping['LOT_SIZE']]["maxQty"])
        base_increment = float(filters[filter_mapping['LOT_SIZE']]["stepSize"])

        min_market_notational = float(filters[filter_mapping['MIN_NOTIONAL']]['minNotional'])
        max_market_notational = 92233720368.547752  # For some reason equal to the first *11 digits* of 2^63 then
                                                    # it gets weird past the decimal

        # Must test both nowadays
        max_orders = int(filters[filter_mapping['MAX_NUM_ORDERS']]["maxNumOrders"])

        if percent_min_price < hard_min_price:
            min_price = hard_min_price
        else:
            min_price = percent_min_price

        if percent_max_price > hard_max_price:
            max_price = hard_max_price
        else:
            max_price = percent_max_price

        return {
            "symbol": symbol,
            "base_asset": symbol_data["baseAsset"],
            "quote_asset": symbol_data["quoteAsset"],
            "max_orders": max_orders,
            "limit_order": {
                "base_min_size": min_quantity,  # Minimum size to buy
                "base_max_size": max_quantity,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment for the base_asset.

                "price_increment": quote_increment,

                "min_price": min_price,
                "max_price": max_price,
            },
            'market_order': {
                "fractionable": True,

                "base_min_size": min_quantity,  # Minimum size to buy
                "base_max_size": max_quantity,  # Maximum size to buy
                "base_increment": base_increment,  # Specifies the minimum increment for the base_asset.

                "quote_increment": quote_increment,  # Specifies the min order price as well as the price increment.
                "buy": {
                    "min_funds": min_market_notational,
                    "max_funds": max_market_notational,
                },
                "sell": {
                    "min_funds": min_market_notational,
                    "max_funds": max_market_notational,
                },
            },
            "exchange_specific": {
                'limit_multiplier_up': multiplier_up,
                'limit_multiplier_down': multiplier_down
            }
        }

    def get_price(self, symbol) -> float:
        """
        Returns just the price of an asset.
        """
        symbol = utils.to_exchange_symbol(symbol, "binance")
        response = self.calls.get_symbol_ticker(symbol=symbol)
        return float(response['price'])



"""
    This support calls to the Coinbase Pro API
    Copyright (c) 2017 Daniel Paquin, modified by Emerson Dove

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.
"""

import base64
import hashlib
import hmac
import json
import time

import requests
from requests.auth import AuthBase

# Create custom authentication for Exchange
from blankly.utils.utils import info_print


class CoinbaseExchangeAuth(AuthBase):
    # Provided by CBPro: https://docs.pro.coinbase.com/#signing-a-message
    def __init__(self, api_key, api_secret, api_pass):
        self.api_key = api_key
        self.secret_key = api_secret
        self.passphrase = api_pass

    def __call__(self, request):
        timestamp = str(time.time())
        message = ''.join([timestamp, request.method,
                           request.path_url, (request.body or '')])
        request.headers.update(get_auth_headers(timestamp, message,
                                                self.api_key,
                                                self.secret_key,
                                                self.passphrase))
        return request


def get_auth_headers(timestamp, message, api_key, secret_key, passphrase):
    message = message.encode('ascii')
    hmac_key = base64.b64decode(secret_key)
    signature = hmac.new(hmac_key, message, hashlib.sha256)
    signature_b64 = base64.b64encode(signature.digest()).decode('utf-8')
    return {
        'Content-Type': 'Application/JSON',
        'CB-ACCESS-SIGN': signature_b64,
        'CB-ACCESS-TIMESTAMP': timestamp,
        'CB-ACCESS-KEY': api_key,
        'CB-ACCESS-PASSPHRASE': passphrase
    }


class API:
    def __init__(self, api_key: str, api_secret: str, api_pass: str, api_url: str = 'https://api.pro.coinbase.com/'):
        self.__auth = CoinbaseExchangeAuth(api_key, api_secret, api_pass)
        self.__api_url = api_url
        self.session = requests.Session()

    """
    Public Client Calls
    """

    """ Full interface support """

    def get_products(self):
        """Get a list of available currency pairs for trading.

        Returns:
            list: Info about all currency pairs. Example::
                [
                    {
                        "id": "BTC-USD",
                        "display_name": "BTC/USD",
                        "base_currency": "BTC",
                        "quote_currency": "USD",
                        "base_min_size": "0.01",
                        "base_max_size": "10000.00",
                        "quote_increment": "0.01"
                    }
                ]
        """
        return requests.get(self.__api_url + 'products', auth=self.__auth).json()

    def get_product_order_book(self, product_id, level=1):
        """Get a list of open orders for a product.

        The amount of detail shown can be customized with the `level`
        parameter:
        * 1: Only the best bid and ask
        * 2: Top 50 bids and asks (aggregated)
        * 3: Full order book (non aggregated)

        Level 1 and Level 2 are recommended for polling. For the most
        up-to-date data, consider using the websocket stream.

        **Caution**: Level 3 is only recommended for users wishing to
        maintain a full real-time order book using the websocket
        stream. Abuse of Level 3 via polling will cause your access to
        be limited or blocked.

        Args:
            product_id (str): Product
            level (Optional[int]): Order book level (1, 2, or 3).
                Default is 1.

        Returns:
            dict: Order book. Example for level 1::
                {
                    "sequence": "3",
                    "bids": [
                        [ price, size, num-orders ],
                    ],
                    "asks": [
                        [ price, size, num-orders ],
                    ]
                }

        """
        if level == 3:
            info_print("Abuse of polling at level 3 can result in a block. Consider using the websocket.")

        params = {'level': level}
        return requests.get(self.__api_url + "products/{}/book".format(product_id), params=params).json()

    """ PAGINATED """ """ Full interface support """

    def get_product_trades(self, product_id, before='', after='', limit=None, result=None):
        """List the latest trades for a product.

        This method returns a generator which may make multiple HTTP requests
        while iterating through it.

        Args:
             product_id (str): Product
             before (Optional[str]): start time in ISO 8601
             after (Optional[str]): end time in ISO 8601
             limit (Optional[int]): the desired number of trades (can be more than 100,
                          automatically paginated)
             result (Optional[list]): list of results that is used for the pagination
        Returns:
             list: Latest trades. Example::
                 [{
                     "time": "2014-11-07T22:19:28.578544Z",
                     "trade_id": 74,
                     "price": "10.00000000",
                     "size": "0.01000000",
                     "side": "buy"
                 }, {
                     "time": "2014-11-07T01:08:43.642366Z",
                     "trade_id": 73,
                     "price": "100.00000000",
                     "size": "0.01000000",
                     "side": "sell"
         }]
        """
        return self._send_paginated_message('/products/{}/trades'
                                            .format(product_id))

    """ Full interface support """

    def get_product_historic_rates(self, product_id, start=None, end=None, granularity=None):
        """Historic rates for a product.

        Rates are returned in grouped buckets based on requested
        `granularity`. If start, end, and granularity aren't provided,
        the exchange will assume some (currently unknown) default values.

        Historical rate data may be incomplete. No data is published for
        intervals where there are no ticks.

        **Caution**: Historical rates should not be polled frequently.
        If you need real-time information, use the trade and book
        endpoints along with the websocket feed.

        The maximum number of data points for a single request is 200
        candles. If your selection of start/end time and granularity
        will result in more than 200 data points, your request will be
        rejected. If you wish to retrieve fine granularity data over a
        larger time range, you will need to make multiple requests with
        new start/end ranges.

        Args:
            product_id (str): Product
            start (Optional[str]): Start time in ISO 8601
            end (Optional[str]): End time in ISO 8601
            granularity (Optional[int]): Desired time slice in seconds

        Returns:
            list: Historic candle data. Example:
                [
                    [ time, low, high, open, close, volume ],
                    [ 1415398768, 0.32, 4.2, 0.35, 4.2, 12.3 ],
                    ...
                ]

        """
        params = {}
        if start is not None:
            params['start'] = start
        if end is not None:
            params['end'] = end

        params['granularity'] = granularity

        return requests.get(self.__api_url + 'products/{}/candles'.format(product_id), params=params).json()

    def get_product_24hr_stats(self, product_id):
        """Get 24 hr stats for the product.

        Args:
            product_id (str): Product

        Returns:
            dict: 24 hour stats. Volume is in base currency units.
                Open, high, low are in quote currency units. Example::
                    {
                        "open": "34.19000000",
                        "high": "95.70000000",
                        "low": "7.06000000",
                        "volume": "2.41000000"
                    }

        """
        return requests.get(self.__api_url + 'products/{}/stats'.format(product_id), auth=self.__auth).json()

    """ Full interface support """

    def get_currencies(self):
        """List known currencies.

        Returns:
            list: List of currencies. Example::
                [{
                    "id": "BTC",
                    "name": "Bitcoin",
                    "min_size": "0.00000001"
                }, {
                    "id": "USD",
                    "name": "United States Dollar",
                    "min_size": "0.01000000"
                }]

        """
        return requests.get(self.__api_url + 'currencies', auth=self.__auth).json()

    def get_time(self):
        """Get the API server time.

        Returns:
            dict: Server time in ISO and epoch format (decimal seconds
                since Unix epoch). Example::
                    {
                        "iso": "2015-01-07T23:47:25.201Z",
                        "epoch": 1420674445.201
                    }

        """
        return requests.get(self.__api_url + 'time', auth=self.__auth).json()

    """
    Private API Calls
    """

    def get_accounts(self):
        """ Get a list of trading all accounts.

        When you place an order, the funds for the order are placed on
        hold. They cannot be used for other orders or withdrawn. Funds
        will remain on hold until the order is filled or canceled. The
        funds on hold for each account will be specified.

        Returns:
            list: Info about all accounts. Example::
                [
                    {
                        "id": "71452118-efc7-4cc4-8780-a5e22d4baa53",
                        "currency": "BTC",
                        "balance": "0.0000000000000000",
                        "available": "0.0000000000000000",
                        "hold": "0.0000000000000000",
                        "profile_id": "75da88c5-05bf-4f54-bc85-5c775bd68254"
                    },
                    {
                        ...
                    }
                ]

        * Additional info included in response for margin accounts.
        """
        return requests.get(self.__api_url + 'accounts', auth=self.__auth).json()

    """ Full interface support """

    def get_account(self, account_id):
        """ Get information for a single account.

        Use this endpoint when you know the account_id.

        Args:
            account_id (str): Account id for account you want to get.

        Returns:
            dict: Account information. Example::
                {
                    "id": "a1b2c3d4",
                    "balance": "1.100",
                    "holds": "0.100",
                    "available": "1.00",
                    "currency": "USD"
                }
        """
        return requests.get(self.__api_url + 'accounts/' + account_id, auth=self.__auth).json()

    """ PAGINATED """

    def get_account_history(self, account_id, **kwargs):
        """ List account activity. Account activity either increases or
        decreases your account balance.

        Entry type indicates the reason for the account change.
        * transfer:	Funds moved to/from Coinbase to cbpro
        * match:	Funds moved as a result of a trade
        * fee:	    Fee as a result of a trade
        * rebate:   Fee rebate as per our fee schedule

        If an entry is the result of a trade (match, fee), the details
        field will contain additional information about the trade.

        Args:
            account_id (str): Account id to get history of.
            kwargs (dict): Additional HTTP request parameters.

        Returns:
            list: History information for the account. Example::
                [
                    {
                        "id": "100",
                        "created_at": "2014-11-07T08:19:27.028459Z",
                        "amount": "0.001",
                        "balance": "239.669",
                        "type": "fee",
                        "details": {
                            "order_id": "d50ec984-77a8-460a-b958-66f114b0de9b",
                            "trade_id": "74",
                            "product_id": "BTC-USD"
                        }
                    },
                    {
                        ...
                    }
                ]
        """
        endpoint = '/accounts/{}/ledger'.format(account_id)
        return self._send_paginated_message(endpoint, params=kwargs)

    """ PAGINATED """

    def get_account_holds(self, account_id, **kwargs):
        """ Get holds on an account.

        This method returns a generator which may make multiple HTTP requests
        while iterating through it.

        Holds are placed on an account for active orders or
        pending withdraw requests.

        As an order is filled, the hold amount is updated. If an order
        is canceled, any remaining hold is removed. For a withdraw, once
        it is completed, the hold is removed.

        The `type` field will indicate why the hold exists. The hold
        type is 'order' for holds related to open orders and 'transfer'
        for holds related to a withdraw.

        The `ref` field contains the id of the order or transfer which
        created the hold.

        Args:
            account_id (str): Account id to get holds of.
            kwargs (dict): Additional HTTP request parameters.

        Returns:
            generator(list): Hold information for the account. Example::
                [
                    {
                        "id": "82dcd140-c3c7-4507-8de4-2c529cd1a28f",
                        "account_id": "e0b3f39a-183d-453e-b754-0c13e5bab0b3",
                        "created_at": "2014-11-06T10:34:47.123456Z",
                        "updated_at": "2014-11-06T10:40:47.123456Z",
                        "amount": "4.23",
                        "type": "order",
                        "ref": "0a205de4-dd35-4370-a285-fe8fc375a273",
                    },
                    {
                    ...
                    }
                ]

        """
        endpoint = '/accounts/{}/holds'.format(account_id)
        return self._send_paginated_message(endpoint, params=kwargs)

    """
    Buy & sell
    """

    def place_order(self, product_id, side, order_type, **kwargs):
        """ Place an order.

        The three order types (limit, market, and stop) can be placed using this
        method. Specific methods are provided for each order type, but if a
        more generic interface is desired this method is available.

        Args:
            product_id (str): Product to order (eg. 'BTC-USD')
            side (str): Order side ('buy' or 'sell)
            order_type (str): Order type ('limit', 'market', or 'stop')
            **client_oid (str): Order ID selected by you to identify your order.
                This should be a UUID, which will be broadcast in the public
                feed for `received` messages.
            **stp (str): Self-trade prevention flag. cbpro doesn't allow self-
                trading. This behavior can be modified with this flag.
                Options:
                'dc'	Decrease and Cancel (default)
                'co'	Cancel oldest
                'cn'	Cancel newest
                'cb'	Cancel both
            **overdraft_enabled (Optional[bool]): If true funding above and
                beyond the account balance will be provided by margin, as
                necessary.
            **funding_amount (Optional[Decimal]): Amount of margin funding to be
                provided for the order. Mutually exclusive with
                `overdraft_enabled`.
            **kwargs: Additional arguments can be specified for different order
                types. See the limit/market/stop order methods for details.

        Returns:
            dict: Order details. Example::
            {
                "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
                "price": "0.10000000",
                "size": "0.01000000",
                "product_id": "BTC-USD",
                "side": "buy",
                "stp": "dc",
                "type": "limit",
                "time_in_force": "GTC",
                "post_only": false,
                "created_at": "2016-12-08T20:02:28.53864Z",
                "fill_fees": "0.0000000000000000",
                "filled_size": "0.00000000",
                "executed_value": "0.0000000000000000",
                "status": "pending",
                "settled": false
            }

        """
        # Margin parameter checks
        if kwargs.get('overdraft_enabled') is not None and \
                kwargs.get('funding_amount') is not None:
            raise ValueError('Margin funding must be specified through use of '
                             'overdraft or by setting a funding amount, but not'
                             ' both')

        # Limit order checks
        if order_type == 'limit':
            if kwargs.get('cancel_after') is not None and \
                    kwargs.get('time_in_force') != 'GTT':
                raise ValueError('May only specify a cancel period when time '
                                 'in_force is `GTT`')
            if kwargs.get('post_only') is not None and kwargs.get('time_in_force') in \
                    ['IOC', 'FOK']:
                raise ValueError('post_only is invalid when time in force is '
                                 '`IOC` or `FOK`')

        # Market and stop order checks
        if order_type == 'market' or order_type == 'stop':
            if not (kwargs.get('size') is None) ^ (kwargs.get('funds') is None):
                raise ValueError('Either `size` or `funds` must be specified '
                                 'for market/stop orders (but not both).')

        # Build params dict
        params = {'product_id': product_id,
                  'side': side,
                  'type': order_type}
        params.update(kwargs)
        return requests.post(self.__api_url + 'orders', data=json.dumps(params), auth=self.__auth).json()

    def place_limit_order(self, product_id, side, price, size,
                          client_oid=None,
                          stp=None,
                          time_in_force=None,
                          cancel_after=None,
                          post_only=None,
                          overdraft_enabled=None,
                          funding_amount=None):
        """Place a limit order.

        Args:
            product_id (str): Product to order (eg. 'BTC-USD')
            side (str): Order side ('buy' or 'sell)
            price (Decimal): Price per cryptocurrency
            size (Decimal): Amount of cryptocurrency to buy or sell
            client_oid (Optional[str]): User-specified Order ID
            stp (Optional[str]): Self-trade prevention flag. See `place_order`
                for details.
            time_in_force (Optional[str]): Time in force. Options:
                'GTC'   Good till canceled
                'GTT'   Good till time (set by `cancel_after`)
                'IOC'   Immediate or cancel
                'FOK'   Fill or kill
            cancel_after (Optional[str]): Cancel after this period for 'GTT'
                orders. Options are 'min', 'hour', or 'day'.
            post_only (Optional[bool]): Indicates that the order should only
                make liquidity. If any part of the order results in taking
                liquidity, the order will be rejected and no part of it will
                execute.
            overdraft_enabled (Optional[bool]): If true funding above and
                beyond the account balance will be provided by margin, as
                necessary.
            funding_amount (Optional[Decimal]): Amount of margin funding to be
                provided for the order. Mutually exclusive with
                `overdraft_enabled`.

        Returns:
            dict: Order details. See `place_order` for example.

        """
        params = {'product_id': product_id,
                  'side': side,
                  'order_type': 'limit',
                  'price': price,
                  'size': size,
                  'client_oid': client_oid,
                  'stp': stp,
                  'time_in_force': time_in_force,
                  'cancel_after': cancel_after,
                  'post_only': post_only,
                  'overdraft_enabled': overdraft_enabled,
                  'funding_amount': funding_amount}
        params = dict((k, v) for k, v in params.items() if v is not None)

        return self.place_order(**params)

    def place_market_order(self, product_id, side, size=None, funds=None,
                           client_oid=None,
                           stp=None,
                           overdraft_enabled=None,
                           funding_amount=None):
        """ Place market order.

        Args:
            product_id (str): Product to order (eg. 'BTC-USD')
            side (str): Order side ('buy' or 'sell)
            size (Optional[Decimal]): Desired amount in crypto. Specify this or
                `funds`.
            funds (Optional[Decimal]): Desired amount of quote currency to use.
                Specify this or `size`.
            client_oid (Optional[str]): User-specified Order ID
            stp (Optional[str]): Self-trade prevention flag. See `place_order`
                for details.
            overdraft_enabled (Optional[bool]): If true funding above and
                beyond the account balance will be provided by margin, as
                necessary.
            funding_amount (Optional[Decimal]): Amount of margin funding to be
                provided for the order. Mutually exclusive with
                `overdraft_enabled`.

        Returns:
            dict: Order details. See `place_order` for example.

        """
        params = {'product_id': product_id,
                  'side': side,
                  'order_type': 'market',
                  'size': size,
                  'funds': funds,
                  'client_oid': client_oid,
                  'stp': stp,
                  'overdraft_enabled': overdraft_enabled,
                  'funding_amount': funding_amount}
        params = dict((k, v) for k, v in params.items() if v is not None)
        return self.place_order(**params)

    def place_stop_order(self, product_id, side, price, size=None, funds=None,
                         client_oid=None,
                         stp=None,
                         overdraft_enabled=None,
                         funding_amount=None):
        """ Place stop order.

        Args:
            product_id (str): Product to order (eg. 'BTC-USD')
            side (str): Order side ('buy' or 'sell)
            price (Decimal): Desired price at which the stop order triggers.
            size (Optional[Decimal]): Desired amount in crypto. Specify this or
                `funds`.
            funds (Optional[Decimal]): Desired amount of quote currency to use.
                Specify this or `size`.
            client_oid (Optional[str]): User-specified Order ID
            stp (Optional[str]): Self-trade prevention flag. See `place_order`
                for details.
            overdraft_enabled (Optional[bool]): If true funding above and
                beyond the account balance will be provided by margin, as
                necessary.
            funding_amount (Optional[Decimal]): Amount of margin funding to be
                provided for the order. Mutually exclusive with
                `overdraft_enabled`.

        Returns:
            dict: Order details. See `place_order` for example.

        """
        params = {'product_id': product_id,
                  'side': side,
                  'stop_price': price,
                  'price': price,
                  'order_type': 'limit',
                  'stop': 'loss',
                  'size': size,
                  'funds': funds,
                  'client_oid': client_oid,
                  'stp': stp,
                  'overdraft_enabled': overdraft_enabled,
                  'funding_amount': funding_amount}
        params = dict((k, v) for k, v in params.items() if v is not None)
        return self.place_order(**params)

    """ Full interface support """

    def cancel_order(self, order_id):
        """ Cancel a previously placed order.

        If the order had no matches during its lifetime its record may
        be purged. This means the order details will not be available
        with get_order(order_id). If the order could not be canceled
        (already filled or previously canceled, etc), then an error
        response will indicate the reason in the message field.

        **Caution**: The order id is the server-assigned order id and
        not the optional client_oid.

        Args:
            order_id (str): The order_id of the order you want to cancel

        Returns:
            list: Containing the order_id of cancelled order. Example::
                [ "c5ab5eae-76be-480e-8961-00792dc7e138" ]

        """
        return requests.delete(self.__api_url + 'orders/' + order_id, auth=self.__auth).json()

    """ PAGINATED """
    """ Full interface support (untested) """

    def get_orders(self, product_id=None, status=None, **kwargs):
        """ List your current open orders.

        This method returns a generator which may make multiple HTTP requests
        while iterating through it.

        Only open or un-settled orders are returned. As soon as an
        order is no longer open and settled, it will no longer appear
        in the default request.

        Orders which are no longer resting on the order book, will be
        marked with the 'done' status. There is a small window between
        an order being 'done' and 'settled'. An order is 'settled' when
        all of the fills have settled and the remaining holds (if any)
        have been removed.

        For high-volume trading it is strongly recommended that you
        maintain your own list of open orders and use one of the
        streaming market data feeds to keep it updated. You should poll
        the open orders endpoint once when you start trading to obtain
        the current state of any open orders.

        Args:
            product_id (Optional[str]): Only list orders for this
                product
            status (Optional[list/str]): Limit list of orders to
                this status or statuses. Passing 'all' returns orders
                of all statuses.
                ** Options: 'open', 'pending', 'active', 'done',
                    'settled'
                ** default: ['open', 'pending', 'active']

        Returns:
            list: Containing information on orders. Example::
                [
                    {
                        "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
                        "price": "0.10000000",
                        "size": "0.01000000",
                        "product_id": "BTC-USD",
                        "side": "buy",
                        "stp": "dc",
                        "type": "limit",
                        "time_in_force": "GTC",
                        "post_only": false,
                        "created_at": "2016-12-08T20:02:28.53864Z",
                        "fill_fees": "0.0000000000000000",
                        "filled_size": "0.00000000",
                        "executed_value": "0.0000000000000000",
                        "status": "open",
                        "settled": false
                    },
                    {
                        ...
                    }
                ]

        """
        params = kwargs
        if product_id is not None:
            params['product_id'] = product_id
        if status is not None:
            params['status'] = status
        return self._send_paginated_message('orders', params=params)

    """ Full interface support (untested) """

    def get_order(self, order_id):
        """ Get a single order by order id.

        If the order is canceled the response may have status code 404
        if the order had no matches.

        **Caution**: Open orders may change state between the request
        and the response depending on market conditions.

        Args:
            order_id (str): The order to get information of.

        Returns:
            dict: Containing information on order. Example::
                {
                    "created_at": "2017-06-18T00:27:42.920136Z",
                    "executed_value": "0.0000000000000000",
                    "fill_fees": "0.0000000000000000",
                    "filled_size": "0.00000000",
                    "id": "9456f388-67a9-4316-bad1-330c5353804f",
                    "post_only": true,
                    "price": "1.00000000",
                    "product_id": "BTC-USD",
                    "settled": false,
                    "side": "buy",
                    "size": "1.00000000",
                    "status": "pending",
                    "stp": "dc",
                    "time_in_force": "GTC",
                    "type": "limit"
                }

        """
        return requests.get(self.__api_url + "orders/" + order_id, auth=self.__auth).json()

    """ PAGINATED """

    def get_fills(self, order_id=None, product_id=None, **kwargs):
        """ Get a list of recent fills.

        As of 8/23/18 - Requests without either order_id or product_id
        will be rejected

        This method returns a generator which may make multiple HTTP requests
        while iterating through it.

        Fees are recorded in two stages. Immediately after the matching
        engine completes a match, the fill is inserted into our
        datastore. Once the fill is recorded, a settlement process will
        settle the fill and credit both trading counterparties.

        The 'fee' field indicates the fees charged for this fill.

        The 'liquidity' field indicates if the fill was the result of a
        liquidity provider or liquidity taker. M indicates Maker and T
        indicates Taker.

        Args:
            product_id (str): Limit list to this product_id
            order_id (str): Limit list to this order_id
            kwargs (dict): Additional HTTP request parameters.

        Returns:
            list: Containing information on fills. Example::
                [
                    {
                        "trade_id": 74,
                        "product_id": "BTC-USD",
                        "price": "10.00",
                        "size": "0.01",
                        "order_id": "d50ec984-77a8-460a-b958-66f114b0de9b",
                        "created_at": "2014-11-07T22:19:28.578544Z",
                        "liquidity": "T",
                        "fee": "0.00025",
                        "settled": true,
                        "side": "buy"
                    },
                    {
                        ...
                    }
                ]

        """
        if (product_id is None) and (order_id is None):
            raise ValueError('Either product_id or order_id must be specified.')

        params = {}
        if product_id:
            params['product_id'] = product_id
        if order_id:
            params['order_id'] = order_id
        params.update(kwargs)

        return self._send_paginated_message('/fills', params=params)

    """ Full interface support (tested) """

    def get_fees(self):
        """
            Returns:
            dict: Contains info about fees and volume. Example::
            {
                'maker_fee_rate': '0.0050',
                'taker_fee_rate': '0.0050',
                'usd_volume': '37.69'
            }
        """
        return requests.get(self.__api_url + "fees", auth=self.__auth).json()

    def _send_paginated_message(self, endpoint, params=None):
        """ Send API message that results in a paginated response.

        The paginated responses are abstracted away by making API requests on
        demand as the response is iterated over.

        Paginated API messages support 3 additional parameters: `before`,
        `after`, and `limit`. `before` and `after` are mutually exclusive. To
        use them, supply an index value for that endpoint (the field used for
        indexing varies by endpoint - get_fills() uses 'trade_id', for example).
            `before`: Only get data that occurs more recently than index
            `after`: Only get data that occurs further in the past than index
            `limit`: Set amount of data per HTTP response. Default (and
                maximum) of 100.

        Args:
            endpoint (str): Endpoint (to be added to base URL)
            params (Optional[dict]): HTTP request parameters

        Yields:
            dict: API response objects

        """
        if params is None:
            params = dict()
        while True:
            r = requests.get(self.__api_url + endpoint, params=params, auth=self.__auth, timeout=30)
            # r = self.session.get(url, params=params, auth=self.auth, timeout=30)
            results = r.json()
            for result in results:
                yield result
            # If there are no more pages, we're done. Otherwise update `after`
            # param to get next page.
            # If this request included `before` don't get any more pages - the
            # cbpro API doesn't support multiple pages in that case.
            if not r.headers.get('cb-after') or \
                    params.get('before') is not None:
                break
            else:
                params['after'] = r.headers['cb-after']

    """
    UNDOCUMENTED
    """

    def create_report(self, report_type, start_date, end_date, product_id=None,
                      account_id=None, report_format='pdf', email=None):
        """ Create report of historic information about your account.

        The report will be generated when resources are available. Report status
        can be queried via `get_report(report_id)`.

        Args:
            report_type (str): 'fills' or 'account'
            start_date (str): Starting date for the report in ISO 8601
            end_date (str): Ending date for the report in ISO 8601
            product_id (Optional[str]): ID of the product to generate a fills
                report for. Required if account_type is 'fills'
            account_id (Optional[str]): ID of the account to generate an account
                report for. Required if report_type is 'account'.
            report_format (Optional[str]): 'pdf' or 'csv'. Default is 'pdf'.
            email (Optional[str]): Email address to send the report to.

        Returns:
            dict: Report details. Example::
                {
                    "id": "0428b97b-bec1-429e-a94c-59232926778d",
                    "type": "fills",
                    "status": "pending",
                    "created_at": "2015-01-06T10:34:47.000Z",
                    "completed_at": undefined,
                    "expires_at": "2015-01-13T10:35:47.000Z",
                    "file_url": undefined,
                    "params": {
                        "start_date": "2014-11-01T00:00:00.000Z",
                        "end_date": "2014-11-30T23:59:59.000Z"
                    }
                }

        """
        params = {'type': report_type,
                  'start_date': start_date,
                  'end_date': end_date,
                  'format': report_format}
        if product_id is not None:
            params['product_id'] = product_id
        if account_id is not None:
            params['account_id'] = account_id
        if email is not None:
            params['email'] = email
        return requests.post(self.__api_url + "reports", data=json.dumps(params), auth=self.__auth).json()

    def get_report(self, report_id):
        """ Get report status.

        Use to query a specific report once it has been requested.

        Args:
            report_id (str): Report ID

        Returns:
            dict: Report details, including file url once it is created.

        """
        return requests.get(self.__api_url + "reports/" + report_id, auth=self.__auth).json()

    def get_trailing_volume(self):
        """  Get your 30-day trailing volume for all products.

        This is a cached value that's calculated every day at midnight UTC.

        Returns:
            list: 30-day trailing volumes. Example::
                [
                    {
                        "product_id": "BTC-USD",
                        "exchange_volume": "11800.00000000",
                        "volume": "100.00000000",
                        "recorded_at": "1973-11-29T00:05:01.123456Z"
                    },
                    {
                        ...
                    }
                ]

        """
        return requests.get(self.__api_url + "users/self/trailing-volume", auth=self.__auth).json()

    def get_coinbase_accounts(self):
        """ Get a list of your coinbase accounts.

        Returns:
            list: Coinbase account details.

        """
        return requests.get(self.__api_url + 'coinbase-accounts', auth=self.__auth).json()

    def get_product_ticker(self, product_id):
        """ Get recent market data for a product
        Use of this is discouraged, websocket feeds should be used instead for high speed polling.

        Returns:
            {
                "trade_id": 4729088,
                "price": "333.99",
                "size": "0.193",
                "bid": "333.98",
                "ask": "333.99",
                "volume": "5957.11914015",
                "time": "2015-11-14T20:46:03.511254Z"
            }
        """
        return requests.get(self.__api_url + 'products/' + product_id + '/ticker', auth=self.__auth).json()

# # Create custom authentication for Exchange
# class CoinbaseExchangeAuth(AuthBase):
#     # Provided by CBPro: https://docs.pro.coinbase.com/#signing-a-message
#     def __init__(self, api_key, secret_key, passphrase):
#         self.api_key = api_key
#         self.secret_key = secret_key
#         self.passphrase = passphrase
#
#     def __call__(self, request):
#         timestamp = str(time.time())
#         message = ''.join([timestamp, request.method,
#                            request.path_url, (request.body or '')])
#         request.headers.update(get_auth_headers(timestamp, message,
#                                                 self.api_key,
#                                                 self.secret_key,
#                                                 self.passphrase))
#         return request
#
#
# def get_auth_headers(timestamp, message, api_key, secret_key, passphrase):
#     message = message.encode('ascii')
#     hmac_key = base64.b64decode(secret_key)
#     signature = hmac.new(hmac_key, message, hashlib.sha256)
#     signature_b64 = base64.b64encode(signature.digest()).decode('utf-8')
#     return {
#         'Content-Type': 'Application/JSON',
#         'CB-ACCESS-SIGN': signature_b64,
#         'CB-ACCESS-TIMESTAMP': timestamp,
#         'CB-ACCESS-KEY': api_key,
#         'CB-ACCESS-PASSPHRASE': passphrase
#     }
#
#
# class API:
#     def __init__(self, API_KEY, API_SECRET, API_PASS, API_URL='https://api.pro.coinbase.com/'):
#         self.__auth = CoinbaseExchangeAuth(API_KEY, API_SECRET, API_PASS)
#         self.__api_url = API_URL
#         self.__Utils = blankly.Utils.Utils()
#
#     def get_portfolio(self, currency=None, show=False):
#         output = requests.get(self.__api_url + 'accounts', auth=self.__auth).json()
#         if show:
#             self.__Utils.printJSON(output)
#
#         if currency is not None:
#             for i in range(len(output)):
#                 if output[i]["currency"] == currency:
#                     out = output[i]
#                     if show:
#                         print(out)
#                     return out
#             raise Exception("Currency not found")
#         return output
#
#     @DeprecationWarning
#     def getAccountInfo(self, currency, property=None, show=False):
#         accounts = self.get_portfolio()
#         if property == None:
#             for i in range(len(accounts)):
#                 if accounts[i]["currency"] == currency:
#                     out = accounts[i]
#                     if show:
#                         print(out)
#                     return out
#         else:
#             for i in range(len(accounts)):
#                 if accounts[i]["currency"] == currency:
#                     out = accounts[i][property]
#                     if show:
#                         print(out)
#                     return out
#
#     """
#     Example placeOrder response (this is a limit):
#     {
#       "status": "pending",
#       "created_at": "2021-01-10T04:39:35.96959Z",
#       "post_only": false,
#       "product_id": "BTC-USD",
#       "fill_fees": "0",
#       "settled": false,
#       "price": "15000",
#       "executed_value": "0",
#       "id": "8e102545-a103-42e0-917a-933a95ecf65b",
#       "time_in_force": "GTC",
#       "stp": "dc",
#       "filled_size": "0",
#       "type": "limit",
#       "side": "buy",
#       "size": "0.001"
#     }
#     """
#
#     """
#     This should be spawned entirely through the Exchange class
#     """
#
#     def placeOrder(self, order, show=False):
#         output = requests.post(self.__api_url + 'orders', json=order, auth=self.__auth)
#
#         if (str(output) == "<Response [400]>"):
#             print(output)
#             self.__Utils.printJSON(output)
#             raise Exception("Invalid Request")
#
#         if show:
#             self.__Utils.printJSON(output)
#         output = output.json()
#         return output
#         # try:
#         #     exchangeLog = Exchange.Exchange(order["side"], order["size"], ticker, output, self, limit=order["price"])
#         # except Exception as e:
#         #     exchangeLog = Exchange.Exchange(order["side"], order["size"], ticker, self, output)
#
#     def getCoinInfo(self, coinID, show=False):
#         output = requests.get(self.__api_url + 'currencies/' + coinID, auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getOpenOrders(self, show=False):
#         output = requests.get(self.__api_url + "orders", auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def deleteOrder(self, id, show=False):
#         output = requests.delete(self.__api_url + "orders/" + id, auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     """ Current maker & taker fee rates as well as your 30-day trading volume """
#
#     def getFees(self, show=False):
#         output = requests.get(self.__api_url + "fees", auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getPriceData(self, timeStart, timeStop, granularity, id, show=False):
#         start = DT.datetime.utcfromtimestamp(timeStart).isoformat()
#         stop = DT.datetime.utcfromtimestamp(timeStop).isoformat()
#         query = {
#             "start": start,
#             "end": stop,
#             "granularity": granularity
#         }
#         response = requests.get(self.__api_url + "products/" + id + "/candles", auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(response)
#         return response
#
#     def getPortfolios(self, show=False):
#         output = requests.get(self.__api_url + "profiles", auth=self.__auth)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getProductData(self, product_id, show=False):
#         output = requests.get(self.__api_url + "products/" + product_id)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getProductOrderBook(self, product_id, show=False):
#         output = requests.get(self.__api_url + "products/" + product_id + "/book")
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getTrades(self, product_id, show=False):
#         output = requests.get(self.__api_url + "products/" + product_id + "/trades")
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getCurrencies(self, id=None, show=False):
#         if id == None:
#             output = requests.get(self.__api_url + "currencies")
#         else:
#             output = requests.get(self.__api_url + "currencies/" + id)
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()
#
#     def getTime(self, show=False):
#         output = requests.get(self.__api_url + "time")
#         if show:
#             self.__Utils.printJSON(output)
#         return output.json()



"""
    Metrics wrapper for backtesting
    Copyright (C) 2021  Brandon Fan, Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import blankly.metrics as metrics
from blankly.utils.time_builder import build_year


def periods_per_year(period: int) -> float:
    """
    Find how many trading periods occur within a trading year

    Args:
        period: the number of seconds in each trading period
    Returns:
        number of trading periods in each year
        
    """
    # TODO: update logic to account for strategies that do not hold overnight
    # 24 hours becomes 6.5 (hours the market is open) AND the risk_free_rate
    # goes to zero
    trading_seconds_per_year = 252.0 * 24.0 * 60.0 * 60.0
    ppy = trading_seconds_per_year / period
    return ppy


def cagr(backtest_data):
    account_values = backtest_data['resampled_account_value']
    years = (account_values['time'].iloc[-1] - account_values['time'].iloc[0]) / build_year()
    return round(metrics.cagr(account_values['value'].iloc[0], account_values['value'].iloc[-1], years), 2) * 100


def cum_returns(backtest_data):
    account_values = backtest_data['resampled_account_value']
    return round(metrics.cum_returns(account_values['value'][0], account_values['value'].iloc[-1]), 2) * 100


def sortino(backtest_data, trading_period=86400, risk_free_rate=0):
    returns = backtest_data['returns']['value']
    ppy = periods_per_year(trading_period)
    return round(metrics.sortino(returns, ppy, risk_free_rate), 2)


def sharpe(backtest_data, trading_period=86400, risk_free_rate=0):
    returns = backtest_data['returns']['value']
    ppy = periods_per_year(trading_period)
    return round(metrics.sharpe(returns, ppy, risk_free_rate), 2)


def calmar(backtest_data, trading_period=86400):
    returns = backtest_data['returns']['value']
    ppy = periods_per_year(trading_period)
    return round(metrics.calmar(returns, ppy), 2)


def volatility(backtest_data, trading_period=86400):
    returns = backtest_data['returns']['value']
    ppy = periods_per_year(trading_period)
    return round(metrics.volatility(returns, ppy), 2)


def variance(backtest_data, trading_period=86400):
    returns = backtest_data['returns']['value']
    ppy = periods_per_year(trading_period)
    return round(100.0 * metrics.variance(returns, ppy), 2)


def beta(backtest_data, trading_period=86400):
    # Drop the first index because it is NaN
    returns = backtest_data['returns']['value'][1:]
    benchmark = backtest_data['benchmark_returns']['value'][1:]
    ppy = periods_per_year(trading_period)
    return round(100.0 * metrics.beta(returns, benchmark, ppy), 2)


def var(backtest_data):
    returns = backtest_data['returns']['value']
    account_values = backtest_data['resampled_account_value']
    return round(metrics.var(account_values['value'][0], returns, 0.95), 2)


def cvar(backtest_data):
    returns = backtest_data['returns']['value']
    account_values = backtest_data['resampled_account_value']
    return round(metrics.cvar(account_values['value'][0], returns, 0.95), 2)


def max_drawdown(backtest_data):
    values = backtest_data['returns']['value']
    return abs(round(metrics.max_drawdown(values), 2)) * 100



"""
    Class to manage a limit order lifecycle - a way to determine individual information about each buy and sell
    Copyright (C) 2022 Matias Kotlik, Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from blankly.exchanges.orders.order import Order


class TakeProfitOrder(Order):
    """
    (Buying or selling (string), amount in currency (BTC/XLM), ticker object (so we can get time and value),
    limit if there is one)
    """

    def __init__(self, order, response, interface):
        self.Interface = interface
        self.__exchange = self.Interface.get_exchange_type()
        self.__order = order
        self.__response = response
        self.__homogenized_result = None
        super().__init__(self.__response, self.__order, self.Interface)

    def get_price(self) -> float:
        """
        Get the price that the order was set at. For limits this will be user-specified, for markets this will
        be market price.
        """
        return self.__response["price"]

    def get_time_in_force(self) -> str:
        """
        Get the exchange's set time_in_force value.
        """
        return self.__response["time_in_force"]

    def __str__(self):
        return_string = super().__str__()

        return_string = self.add_new_line(return_string, "Take Profit Order Parameters: ")

        return_string = self.add_new_line(return_string, "Time In Force: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_time_in_force())

        return_string = self.add_new_line(return_string, "Price: ", newline=False)
        return_string = self.add_new_line(return_string, self.get_price())

        return return_string

    # override this here to homogenize these strings
    def get_type(self) -> str:
        return 'take_profit'
